/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

/* Constants for character set conversion. */
#define BYTEMASK_3   0xFFFFF800UL   /* mask to test for 3-byte UTF-8 representation required */
#define BYTEMASK_2   0xFFFFFF80UL   /* mask to test for 2-byte UTF-8 representation required */

#define BYTEU8_LEN6  0xFC           /* length prefix for 6-byte UTF-8 sequence */
#define BYTEU8_LEN5  0xF8           /* length prefix for 5-byte UTF-8 sequence */
#define BYTEU8_LEN4  0xF0           /* length prefix for 4-byte UTF-8 sequence */
#define BYTEU8_LEN3  0xE0           /* length prefix for 3-byte UTF-8 sequence */
#define BYTEU8_LEN2  0xC0           /* length prefix for 2-byte UTF-8 sequence */
#define BYTEU8_PFIX  0x80           /* prefix for supplemental bytes of a multibyte sequence */

#define BYTETEST_1   0x80           /* test mask for 1-byte UTF-8 sequence */
#define BYTETEST_PF  0xC0           /* test mask for second/subsequent bytes of multibyte */
#define BYTETEST_2   0xE0           /* test mask for 2-byte UTF-8 sequence */
#define BYTETEST_3   0xF0           /* test mask for 3-byte UTF-8 sequence */
#define BYTETEST_4   0xF8           /* test mask for 4-byte UTF-8 sequence */
#define BYTETEST_5   0xFC           /* test mask for 5-byte UTF-8 sequence */
#define BYTETEST_6   0xFE           /* test mask for 6-byte UTF-8 sequence */

#define LOBITS_1     0x7F           /* low-bit mask for 1-byte UTF-8 sequence */
#define LOBITS_PFX   0x3F           /* low-bit mask for supp. bytes of multibyte sequence */
#define LOBITS_2     0x1F           /* low-bit mask for 2-byte UTF-8 sequence */
#define LOBITS_3     0x0F           /* low-bit mask for 3-byte UTF-8 sequence */
#define LOBITS_4     0x07           /* low-bit mask for 4-byte UTF-8 sequence */
#define LOBITS_5     0x03           /* low-bit mask for 5-byte UTF-8 sequence */
#define LOBITS_6     0x01           /* low-bit mask for 6-byte UTF-8 sequence */

#define SAFE2_MASK   0xFE           /* test for safe encoding of 2-byte sequences (mask) */
#define SAFE2_TEST   0xC0           /* test for safe encoding of 2-byte sequences (value) */
#define SAFE3A_TEST  0xE0           /* test for safe encoding of 3-byte sequences (value 1) */
#define SAFE3B_MASK  0xE0           /* test for safe encoding of 3-byte sequences (mask 2) */
#define SAFE3B_TEST  0x80           /* test for safe encoding of 3-byte sequences (value 2) */
#define SAFE4A_TEST  0xF0           /* test for safe encoding of 4-byte sequences (value 1) */
#define SAFE4B_MASK  0xF0           /* test for safe encoding of 4-byte sequences (mask 2) */
#define SAFE4B_TEST  0x80           /* test for safe encoding of 4-byte sequences (value 2) */
#define SAFE5A_TEST  0xF8           /* test for safe encoding of 5-byte sequences (value 1) */
#define SAFE5B_MASK  0xF8           /* test for safe encoding of 5-byte sequences (mask 2) */
#define SAFE5B_TEST  0x80           /* test for safe encoding of 5-byte sequences (value 2) */
#define SAFE6A_TEST  0xFC           /* test for safe encoding of 6-byte sequences (value 1) */
#define SAFE6B_MASK  0xFC           /* test for safe encoding of 6-byte sequences (mask 2) */
#define SAFE6B_TEST  0x80           /* test for safe encoding of 6-byte sequences (value 2) */

#define BITOFS       6              /* bit offset for fragments of UTF-8 data */
#define BITSOFS(x)   ((x) * BITOFS) /* multiples of the bit offset */

#define UCS4SUB      0xFFFDUL       /* UCS-4 substitute character */
#define WINSUB       '\xBF'         /* Windows substitute character */
#define NULCHR       '\0'           /* null character for termination */

/* Conversion table for converting from Windows-1252 to UCS-4 (values 0x80-0x9F) */
static const unsigned long windows_1252_80_9F[] =
{
  0x20ACUL, UCS4SUB,  0x201AUL, 0x0192UL, 0x201EUL, 0x2026UL, 0x2020UL, 0x2021UL,
  0x02C6UL, 0x2030UL, 0x0160UL, 0x2039UL, 0x0152UL, UCS4SUB,  0x017DUL, UCS4SUB,
  UCS4SUB,  0x2018UL, 0x2019UL, 0x201CUL, 0x201DUL, 0x2022UL, 0x2013UL, 0x2014UL,
  0x02DCUL, 0x2122UL, 0x0161UL, 0x203AUL, 0x0153UL, UCS4SUB,  0x017EUL, 0x0178UL
};

/* Conversion tables for converting from UCS-4 to Windows-1252 */
static const char ucs_conv_20xx[] =  /* this is for the page U+20nn */
{
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 00 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, '\x96', '\x97', WINSUB, WINSUB, WINSUB, /* 10 */
  '\x91', '\x92', '\x82', WINSUB, '\x93', '\x94', '\x84', WINSUB,
  '\x86', '\x87', '\x95', WINSUB, WINSUB, WINSUB, '\x85', WINSUB, /* 20 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  '\x89', WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 30 */
  WINSUB, '\x8B', '\x9B', WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 40 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 50 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 60 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 70 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 80 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 90 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* A0 */
  WINSUB, WINSUB, WINSUB, WINSUB, '\x80', WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* B0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* C0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* D0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* E0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* F0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB
};

static const char ucs_conv_01xx[] = /* this is for the page U+01nn */
{
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 00 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 10 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 20 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 30 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 40 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, '\x8C', '\x9C', WINSUB, WINSUB, WINSUB, WINSUB, /* 50 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  '\x8A', '\x9A', WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 60 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 70 */
  '\x9F', WINSUB, WINSUB, WINSUB, WINSUB, '\x8E', '\x9E', WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 80 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, '\x83', WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* 90 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* A0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* B0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* C0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* D0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* E0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB,
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, /* F0 */
  WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB, WINSUB
};

char *convert_windows_to_utf8_len(pool p, const char *windows_str, int len)
/*
    Description:
        Converts a string encoded in the Windows-1252 character set ("Windows Latin-1",
	"code page 1252") to the UTF-8 encoding of UCS-4.  Characters that are
	"undefined" in Windows-1252 are replaced with the "replacement character" of
	UCS-2 (U+FFFD).

    Input:
        Parameters:
	    p - The pool to allocate the returned string from, or NULL to allocate the
	        string from the heap directly.
	    windows_str - The string to convert.
	    len - Length of the string to convert in bytes.

    Output:
        Returns:
	    NULL - Memory failure, or input is NULL.
	    Other - Pointer to the string converted to UTF-8.
*/
{
    const unsigned char *w2_str = (const unsigned char *)windows_str;
    unsigned long *ucs4str;         /* character string in UCS-4 */
    int utflen = 0;                 /* length of string in UTF-8 */
    register int i;                 /* loop counter */
    char *rc;                       /* return from this function */
    unsigned char *fill;            /* fill pointer */

    if (!windows_str || (len<=0))
        return NULL;   /* safety guard */

    /* STEP 1 - Convert Windows-1252 to UCS-4 */
    ucs4str = (unsigned long *) pmalloc(p,len * sizeof(unsigned long));

    for (i=0; i<len; i++)
    { /* most Windows-1252 characters are the same as Latin-1, and hence can go
       * straight through, except for 0x80-0x9F, which go through the conversion table.
       */
        unsigned long tmp = (unsigned long) (w2_str[i]) & 0xFFUL;
        if ((tmp >= 0x80UL) && (tmp <= 0x9FUL))
            ucs4str[i] = windows_1252_80_9F[(int)(tmp - 0x80UL)];
        else
            ucs4str[i] = tmp;
    }

    /* STEP 2 - Determine the length of the UTF-8 string */
    /* This is simplified because all characters are in the BMP, so we will never
     * encode a sequence longer than 3 bytes
     */
    for (i=0; i<len; i++)
    { /* add various byte lengths */
        if ((ucs4str[i] & BYTEMASK_3) != 0)
            utflen += 3;
        else if ((ucs4str[i] & BYTEMASK_2) != 0)
            utflen += 2;
        else
            utflen++;
    }

    /* STEP 3 - Allocate and fill the output string */
    /* This is simplified because all characters are in the BMP, so we will never
     * encode a sequence longer than 3 bytes
     */
    rc = pmalloc(p,(utflen + 1) * sizeof(char));

    if (rc)
    { /* figure out how to fill the UTF-8 string */
        fill = (unsigned char *)rc;
        for (i=0; i<len; i++)
        { /* fill the output array */
            if ((ucs4str[i] & BYTEMASK_3)!=0)
            { /* 3-byte fill sequence */
                *fill++ = (unsigned char)(((ucs4str[i] >> BITSOFS(2)) & LOBITS_3) | BYTEU8_LEN3);
                *fill++ = (unsigned char)(((ucs4str[i] >> BITSOFS(1)) & LOBITS_PFX) | BYTEU8_PFIX);
                *fill++ = (unsigned char)((ucs4str[i] & LOBITS_PFX) | BYTEU8_PFIX);
                
            }
            else if ((ucs4str[i] & BYTEMASK_2)!=0)
            { /* 2-byte fill sequence */
                *fill++ = (unsigned char)(((ucs4str[i] >> BITSOFS(1)) & LOBITS_2) | BYTEU8_LEN2);
                *fill++ = (unsigned char)((ucs4str[i] & LOBITS_PFX) | BYTEU8_PFIX);
            }
            else  /* 1-byte fill sequence */
                *fill++ = (unsigned char)(ucs4str[i] & LOBITS_1);
        }

        *fill = (unsigned char)NULCHR;
    }

    return rc;
}

char *convert_windows_to_utf8(pool p, const char *windows_str)
{
    if (windows_str)
        return convert_windows_to_utf8_len(p,windows_str,strlen(windows_str));
    else
        return NULL;
}

int get_utf8_to_windows_len(const char *utf8_str)
/*
    Description:
        Determines the number of Windows-1252 characters required to hold a conversion
	from the UTF-8 encoded string.

    Input:
        Parameters:
	    utf8_str - The string to get the conversion length of.

    Output:
        Returns:
	    -1 - String format failure, or input is NULL.
	    Other - Number of Windows-1252 characters required to express the
	            UTF-8 string.
*/
{
    const unsigned char *t; /* conversion pointer */
    register int i;         /* loop counter */
    int winlen = 0;         /* length of string in UCS-4 and Windows-1252 */

    if (!utf8_str)
        return -1;

    /* Begin determining the UTF-8 string length, and also error-check the UTF-8 encoding. */
    t = (const unsigned char *)utf8_str;
    while (*t)
    {
        /* determine the "reach" of this byte sequence */
        if ((*t & BYTETEST_1)==0)
            i = 1;
        else if ((*t & BYTETEST_2)==BYTEU8_LEN2)
        { /* 2-byte sequence - check to make sure we have enough bits */
            if ((*t & SAFE2_MASK)==SAFE2_TEST)
            {
                /* we should have been able to encode in 1 byte - this is an error! */
                log_error(ZONE,"UTF-8 HOSATION! excessive encoding in 2-byte sequence");
                return -1;
            }

            i = 2;  /* 2-byte sequence */

        }
        else if ((*t & BYTETEST_3)==BYTEU8_LEN3)
        {
            /* 3-byte sequence - check to make sure we have enough bits */
            if ((*t==SAFE3A_TEST) && ((*(t+1) & SAFE3B_MASK)==SAFE3B_TEST))
            { /* we should have been able to encode in 2 bytes or less - this is an error! */
                log_error(ZONE,"UTF-8 HOSATION! excessive encoding in 3-byte sequence");
                return -1;
            }

            i = 3;  /* 3-byte sequence */

        }
        else if ((*t & BYTETEST_4)==BYTEU8_LEN4)
        { /* 4-byte sequence - check to make sure we have enough bits */
            if ((*t==SAFE4A_TEST) && ((*(t+1) & SAFE4B_MASK)==SAFE4B_TEST))
            { /* we should have been able to encode in 3 bytes or less - this is an error! */
                log_error(ZONE,"UTF-8 HOSATION! excessive encoding in 4-byte sequence");
                return -1;
            }

            i = 4;  /* 4-byte sequence */
        }
        else if ((*t & BYTETEST_5)==BYTEU8_LEN5)
        { /* 5-byte sequence - check to make sure we have enough bits */
            if ((*t==SAFE5A_TEST) && ((*(t+1) & SAFE5B_MASK)==SAFE5B_TEST))
            { /* we should have been able to encode in 4 bytes or less - this is an error! */
                log_error(ZONE,"UTF-8 HOSATION! excessive encoding in 5-byte sequence");
                return -1;
            }

            i = 5;  /* 5-byte sequence */

        }
        else if ((*t & BYTETEST_6)==BYTEU8_LEN6)
        { /* 6-byte sequence - check to make sure we have enough bits */
            if ((*t==SAFE6A_TEST) && ((*(t+1) & SAFE6B_MASK)==SAFE6B_TEST))
            { /* we should have been able to encode in 5 bytes or less - this is an error! */
                log_error(ZONE,"UTF-8 HOSATION! excessive encoding in 6-byte sequence");
                return -1;
            }

            i = 6;  /* 6-byte sequence */

        }
        else
        { /* the client didn't send valid UTF-8! */
            log_error(ZONE,"UTF-8 HOSATION! prefix byte 0x%02X not valid",(unsigned)(*t));
            return -1;
        }

        while (i>0)
        { /* skip ahead in the sequence */
            t++;
            if ((--i)>0)
            { /* make sure the following bytes are in the correct pattern */
                if ((*t & BYTETEST_PF)!=BYTEU8_PFIX)
                { /* the client didn't send valid UTF-8! */
                    log_error(ZONE,"UTF-8 HOSATION! infix byte 0x%02X not valid",(unsigned)(*t));
                    return -1;

                } /* end if (byte not a valid part of a UTF-8 multibyte sequence) */

            }   
        }
    winlen++;
  }

  return winlen;

} /* end get_utf8_to_windows_len */

char *convert_utf8_to_windows(pool p, const char *utf8_str)
/*
    Description:
        Converts a string encoded in the UTF-8 encoding of UCS-4 to the Windows-1252
	encoding ("Windows Latin-1", "code page 1252").  Characters that cannot be
	converted to this encoding are replaced with the inverted question-mark
	character (0xBF in the Windows-1252 charset).

    Input:
        Parameters:
	    p - The pool to allocate the returned string from, or NULL to allocate the
	        string from the heap directly.
	    utf8_str - The string to convert.

    Output:
        Returns:
	    NULL - Memory or string format failure, or input is NULL.
	    Other - Pointer to the string converted to Windows-1252.
*/
{
    unsigned long *ucs4str; /* character string in UCS-4 */
    int winlen = 0;         /* length of string in UCS-4 and Windows-1252 */
    register int i;         /* loop counter */
    char *rc;               /* return from this function */
    const unsigned char *t; /* conversion pointer */

    if (!utf8_str)
        return NULL;

    /* STEP 1 - Determine how long the UCS-4 buffer (and the Windows return string)
     * should be; this also error-checks the string to see if it uses valid UTF-8
     * encoding
     */
    winlen = get_utf8_to_windows_len(utf8_str);
    if (winlen<0)
        return NULL;

    /* STEP 2 - Convert the UTF-8 string to UCS-4 */
    ucs4str = (unsigned long *)malloc(winlen * sizeof(unsigned long));
    if (!ucs4str)
    { /* there was a memory failure here - bogus! */
        log_error(ZONE,"convert_windows_to_utf8: memory error allocating UCS-4 buffer of %d bytes",
                  winlen * sizeof(unsigned long));
        return NULL;
    }

    i = 0;
    t = (const unsigned char *)utf8_str;
    while (*t)
    {
        /* grab each UCS-4 character from the input in turn */
        unsigned long chr = 0;

        if ((*t & BYTETEST_1)==0)
            chr = (unsigned long)(LOBITS_1 & *t++);
        else if ((*t & BYTETEST_2)==BYTEU8_LEN2)
        { /* convert a 2-byte sequence */
            chr = (unsigned long)(LOBITS_2 & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
        }
        else if ((*t & BYTETEST_3)==BYTEU8_LEN3)
        { /* convert a 3-byte sequence */
            chr = (unsigned long)(LOBITS_3 & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
        }
        else if ((*t & BYTETEST_4)==BYTEU8_LEN4)
        {
            /* convert a 4-byte sequence */
            chr = (unsigned long)(LOBITS_4 & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
        }
        else if ((*t & BYTETEST_5)==BYTEU8_LEN5)
        { /* convert a 5-byte sequence */
            chr = (unsigned long)(LOBITS_5 & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            
        } /* end else if */
        else if ((*t & BYTETEST_6)==BYTEU8_LEN6)
        { /* convert a 6-byte sequence */
            chr = (unsigned long)(LOBITS_6 & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
            chr = (chr << BITOFS) | (unsigned long)(LOBITS_PFX & *t++);
        }
        else
        { /* an erroneous sequence - shouldn't we have caught this earlier? */
            log_error(ZONE,"convert_utf8_to_windows: we should have caught this error earlier!");
            return NULL;
        }

        ucs4str[i++] = chr;
    }

    if (i!=winlen)
    { /* sanity check here */
        log_error(ZONE,"SANITY CHECK FAILURE in UTF-8 conversion");
        free(ucs4str);
        return NULL;

    }

    /* STEP 3 - Allocate and fill the output string, converting from
     * UCS-4 to Windows-1252
     */

    rc = pmalloc(p,(winlen + 1) * sizeof(char));

    /* begin filling the output */
    for (i = 0; i < winlen; i++)
    {
        /* convert all the UCS-4 characters to Windows-1252 */
        if (!(ucs4str[i] & 0xFFFF0000))
        {
            /* convert character from BMP */
            unsigned long lobyte = ucs4str[i] & 0xFFUL;
            switch ((ucs4str[i] >> 8) & 0xFFUL)
            {
                /* convert characters here... */
            case 0x00UL:
                /* most characters will come here, just filter out "upper controls" */
                if ((lobyte>=0x80UL) && (lobyte<=0x9FUL))
                    rc[i] = WINSUB;
                else
                    rc[i] = (char)lobyte;
                break;

            case 0x01UL:  /* just use the conversion table */
                rc[i] = ucs_conv_01xx[lobyte];
                break;

            case 0x02UL:  /* just 2 characters to check for here */
                if (lobyte==0xC6UL)
                    rc[i] = '\x88';
                else if (lobyte==0xDCUL)
                    rc[i] = '\x98';
                else
                    rc[i] = WINSUB;  /* huh? */
                break;

            case 0x20UL:  /* just use the conversion table */
                rc[i] = ucs_conv_20xx[lobyte];
                break;

            case 0x21UL:  /* just 1 character to check for here */
                if (lobyte==0x22UL)
                    rc[i] = '\x99';
                else
                    rc[i] = WINSUB;
                break;

            default:  /* huh? */
                rc[i] = WINSUB;
                break;
            }   
        }
        else /* all non-BMP characters get the "huh?" character */
            rc[i] = WINSUB;
    }

    rc[winlen] = NULCHR;  /* null-terminate the output string appropriately */

    free(ucs4str);
    return rc;
}
