/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_contact_load_roster(session s)
{
    xmlnode roster, item;
    jid id;
    UIN_t uin;
    contact c;

    id = it_xdb_id(s->p,s->id,s->from->server);
    roster = xdb_get(s->ti->xc,id,NS_ROSTER);
    if (roster == NULL)
    {
        log_debug(ZONE,"User has no roster");
        return;
    }

    for (item = xmlnode_get_firstchild(roster); item != NULL; item = xmlnode_get_nextsibling(item))
    {
        uin = it_strtouin_safe(xmlnode_get_attrib(item,"jid"));
        if (uin)
        {
            c = it_contact_add(s,uin);
            c->status = STATUS_OFFLINE;
        }
    }

    xmlnode_free(roster);
}

contact it_contact_get(session s, UIN_t uin)
{
    contact c;

    for (c = s->contacts; c != NULL; c = c->next)
        if (c->uin == uin) break;

    return c;
}

contact it_contact_add(session s, UIN_t uin)
{
    contact c;
    pool p;

    p = pool_new();
    c = (contact) pmalloco(p,sizeof(_contact));
    c->p = p;

    c->uin = uin;
    c->status = STATUS_NOT_IN_LIST;
    c->s = s;
    c->addr.sin_family = AF_INET;
    c->tcp_seq = 0xFFFFFFFFUL;
    c->queue = pth_msgport_create("it_contact_queue");
    c->force_server = 1;

    c->next = s->contacts;
    s->contacts = c;

    return c;
}

void it_contact_remove(contact c)
{
    session s = c->s;
    contact cur, prev;

    for (prev = NULL, cur = s->contacts; cur != c; prev = cur, cur = cur->next);

    if (prev)
        prev->next = cur->next;
    else
        s->contacts = cur->next;

    if (c->m)
        it_peer_close(c);

    pool_free(c->p);
}

void it_contact_update_addresses(contact c, unsigned long ip, unsigned short port, unsigned long real_ip, int force_server)
{
    char ip1[20], ip2[20];
    struct in_addr blort;

    blort.s_addr = ip;
    strcpy(ip1,inet_ntoa(blort));

    blort.s_addr = real_ip;
    strcpy(ip2,inet_ntoa(blort));

    memset(&(c->addr),0,sizeof(struct sockaddr_in));
    c->addr.sin_family = AF_INET;
    c->addr.sin_addr.s_addr = ip;
    c->addr.sin_port = port;//htons(port);
    c->addr_real.s_addr = real_ip;
    c->force_server = force_server ? 1 : 0;
    c->port = (int) port;

    log_debug(ZONE,"Contact[%lu], address %s:%d",c->uin,inet_ntoa(c->addr.sin_addr),port);
}

void it_contact_set_status(contact c, icqstatus status)
{
    int delta = (c->status != status);

    c->status = status;
    if (delta)
    {
        if (status == STATUS_OFFLINE && c->m)
            it_peer_close(c);

        it_contact_send_presence(c);
    }
}

void it_contact_send_presence(contact c)
{
    session s = c->s;
    xmlnode pres, x;
    icqstatus status = c->status;
    jid id;

    pres = jutil_presnew(((status <= STATUS_OFFLINE) ? JPACKET__UNAVAILABLE : JPACKET__AVAILABLE),
                        jid_full(s->id),NULL);

    id = it_uin2jid(xmlnode_pool(pres),c->uin,s->from->server);
    xmlnode_put_attrib(pres,"from",jid_full(id));

    switch (status)
    {
    case STATUS_AWAY:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"away",-1);
        break;
    case STATUS_DND:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"dnd",-1);
        break;
    case STATUS_NA:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"xa",-1);
        break;
    case STATUS_OCCUPIED:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"busy",-1);
        break;
    case STATUS_FREE_CHAT:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"chat",-1);
        break;
    default:
        break;
    }

    it_deliver(s->ti,pres);
}

void it_contact_subscribe(void *arg)
{
    contact c = (contact) arg;
    session s = c->s;
    iti ti = s->ti;
    xmlnode x, roster;
    jid id;
    pool p;
    char uin[16];

    x = jutil_presnew(JPACKET__SUBSCRIBED,jid_full(s->id),NULL);
    p = xmlnode_pool(x);
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,c->uin,s->from->server)));

    id = it_xdb_id(p,s->id,s->from->server);
    roster = xdb_get(ti->xc,id,NS_ROSTER);
    if (roster == NULL)
    {
        roster = xmlnode_new_tag("query");
        xmlnode_put_attrib(roster,"xmlns",NS_ROSTER);
    }

    snprintf(uin,15,"%lu",c->uin);

    if (xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p)) == NULL)
    {
        xmlnode_put_attrib(xmlnode_insert_tag(roster,"item"),"jid",uin);
        if (xdb_set(ti->xc,id,NS_ROSTER,roster))
        {
            log_error(s->from->server,"Roster set failed for '%s'",jid_full(s->id));
            xmlnode_free(roster);
        }
    }
    else
        xmlnode_free(roster);

    it_message_send(c,it_message_create(NULL,c->uin,MSGTYPE_CONTACT_ADD,s->nick,s->first,s->last,s->email));

    c->subscribing = 0;
    c->status = STATUS_OFFLINE;

    it_server_send_contact_list(s);

    /* send back the "subscribed" response */
    it_deliver(ti,x);
}

void it_contact_unsubscribe(void *arg)
{
    contact c = (contact) arg;
    session s = c->s;
    iti ti = s->ti;
    xmlnode x;
    jid id;
    xmlnode roster, item;
    pool p;
    char uin[16];

    x = jutil_presnew(JPACKET__SUBSCRIBED,jid_full(s->id),NULL);
    p = xmlnode_pool(x);
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,c->uin,s->from->server)));

    id = it_xdb_id(p,s->id,s->from->server);
    roster = xdb_get(ti->xc,id,NS_ROSTER);

    if (roster != NULL)
    {
        snprintf(uin,15,"%lu",c->uin);
        item = xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p));

        if (item != NULL)
        {
            xmlnode_hide(item);
            if (xdb_set(ti->xc,id,NS_ROSTER,roster))
            {
                log_error(s->from->server,"Roster set failed for '%s'",jid_full(s->id));
                xmlnode_free(roster);
            }
        }
        else
            xmlnode_free(roster);

        it_contact_remove(c);
        it_server_send_contact_list(s);
    }

    it_deliver(ti,x);
}
