/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

static icqbyte *it_laydown_number(icqbyte *buffer, unsigned long value, char term);
static icqbyte *it_laydown_string(pool p, icqbyte *buffer, char *string, char term);

imessage it_message_create(xmlnode x, UIN_t to, unsigned short type, ...)
{
    imessage m;
    pool p;
    va_list argp;
    char *tmp, **parts;
    int total, i;
    int convert = 0;

    p = x ? xmlnode_pool(x) : pool_new();
    m = pmalloco(p,sizeof(_imessage));
    m->x = x;
    m->p = p;
    m->type = type;
    m->to = to;

    switch (type)
    {
    case MSGTYPE_TEXT:
    case MSGTYPE_AUTH_DENY:
        total = 1;
        convert = 1;
        break;
    case MSGTYPE_URL:
        total = 2;
        convert = 1;
        break;
    case MSGTYPE_CONTACT_ADD:
        total = 4;
        break;
    case MSGTYPE_AUTH_REQUEST:
        total = 5;
        break;
    default:
        total = 0;
    }

    log_debug(ZONE,"New message, %d parts",total);

    if (total)
    {
        va_start(argp,type);

        parts = m->parts = pmalloco(p,total * sizeof(char *));
        for (i = 0; i < total; i ++)
        {
            /* do all the string conversions */
            tmp = va_arg(argp,char *);

            if (tmp != NULL)
            {
                tmp = convert ? convert_utf8_to_windows(p,tmp) : pstrdup(p,tmp);
                if (tmp == NULL)
                {
                    log_debug(ZONE,"Error converting to Windows from UTF-8");
                    return NULL;
                }

                parts[i] = it_strrepl(p,tmp,s_LF,s_CRLF);
            }
            else
                parts[i] = "";
        }

        va_end(argp);
    }

    return m;
}

/* Sends a message packet out via the peer connection for this contact if
   possible, or via the server connection if not. */
void it_message_send(contact c, imessage m)
{
    session s = c->s;

    if (s->ti->disable_tcp || c->force_server || c->status < 0 )
    {
        /* we must send through the server */
        if (s->ti->disable_tcp == 0)
            log_debug(ZONE,"force send through server");

        if (it_message_truncate(m))
        {
            log_debug(ZONE,"Message data too long");
            if (m->x)
            {
                jutil_error(m->x,(terror) {501,"Message data too long"});
                it_deliver(s->ti,m->x);
                return;
            }
        }
        else
            it_server_send_message(c->s,m);

        pool_free(m->p);
        return;
    }

    if (c->connected)
    {
        it_peer_send_message(c,m);
        pool_free(m->p);
        return;
    }

    if (c->m == NULL)
        it_peer_connect(c);

    log_debug(ZONE,"Queued");

    pth_msgport_put(c->queue,(pth_message_t *) m);
}

int it_message_size(imessage m)
{
    char **parts = m->parts;
    int rc = 0;

    switch (m->type)
    {
    case MSGTYPE_TEXT:
    case MSGTYPE_AUTH_DENY:
        rc = strlen(parts[0]) + 1;
        break;

    case MSGTYPE_URL:
        rc = strlen(parts[0]) + strlen(parts[1]) + 1;
        break;

    case MSGTYPE_AUTHORIZE:
        rc = sizeof(icqshort);
        break;

    case MSGTYPE_CONTACT_ADD:
    case MSGTYPE_AUTH_REQUEST:
    {
        int i;

        /* nick, first, last, email */
        for (i = 0; i < 4; i++)
            rc += strlen(parts[i]) + 1;

        ++rc;
        if (m->type == MSGTYPE_CONTACT_ADD)
            break;

        rc += strlen(parts[4]) + 2;
        break;
    }

    case MSGTYPE_CONTACTS:
    {
        char num[16];
        int i;

        for (i = 0; parts[i] != NULL; i += 2)
            rc += strlen(parts[i]) + strlen(parts[i + 1]) + 2;

        snprintf(num,16,"%d",i);
        rc = strlen(num) + 1;
        break;
    }

    default:
        log_debug(ZONE,"unknown message type %u",m->type);
        break;
    }

    return rc;
}

icqbyte *it_message_format(imessage m, icqbyte *buffer)
{
    icqbyte *current = buffer;   /* current pointer within the buffer */
    pool p = m->p;
    char **parts = m->parts;
    int i;

    switch (m->type)
    {
    case MSGTYPE_TEXT:       /* ordinary message text */
    case MSGTYPE_AUTH_DENY:  /* authorization denied */
        current = it_laydown_string(p,current,parts[0],'\0');
        break;

    case MSGTYPE_URL:
        current = it_laydown_string(p,current,parts[0],SEPARATOR_CHAR);   /* description */
        current = it_laydown_string(p,current,parts[1],'\0');             /*URL */
        break;

    case MSGTYPE_CONTACT_ADD:
    case MSGTYPE_AUTH_REQUEST:
        for (i = 0; i < 4; i++)
            current = it_laydown_string(p,current,parts[i],SEPARATOR_CHAR);

        if (m->type == MSGTYPE_CONTACT_ADD)
        {
            *current++ = '\0';
            break;
        }

        *current++ = (icqbyte)'0';   /* auth flag */
        *current++ = (icqbyte)SEPARATOR_CHAR;

        /* reason */
        current = it_laydown_string(p,current,parts[4],'\0');
        break;

    case MSGTYPE_AUTHORIZE:
        put_icqshort(current,0,MSG_AUTHORIZE_MAGIC);
        current += sizeof(icqshort);
        break;

    case MSGTYPE_CONTACTS:
    {
        icqbyte *temp;

        temp = current;
        current = it_laydown_number(current,(unsigned long) 0,SEPARATOR_CHAR);

        for (i = 0; parts[i] != NULL; i += 2)
        {
            /* add each entry in turn */
            current = it_laydown_string(p,current,parts[i],SEPARATOR_CHAR);
            current = it_laydown_string(p,current,parts[i + 1],SEPARATOR_CHAR);
        }

        it_laydown_number(temp,(unsigned long) i/2,SEPARATOR_CHAR);
        *current++ = (icqbyte)'\0';  /* add the final terminator */
        break;
    }

    default:
        log_debug(ZONE,"unknown message type %u",m->type);
        return NULL;
    }

    return current;
}

static icqbyte *it_laydown_string(pool p, icqbyte *buffer, char *str, char term)
{
    if (str && *str)
    {
        int len;

        /* Copy the string down and terminate it appropriately. */
        len = strlen(str);
        memcpy(buffer,str,len);
        buffer += len;
    }

    *buffer++ = (icqbyte)term;
    return buffer;
}

static icqbyte *it_laydown_number(icqbyte *buffer, unsigned long value, char term)
{
    char cvtbuf[16];
    int len;

    snprintf(cvtbuf,16,"%lu",value);
    len = strlen(cvtbuf);
    memcpy(buffer,cvtbuf,len);
    buffer += len;
    *buffer++ = (icqbyte)term;

    return buffer;
}

int it_message_truncate(imessage m)
{
    char **parts = m->parts;
    char *tmp;
    int maxlen;

    switch (m->type)
    {
    case MSGTYPE_TEXT:            /* normal text message */
    case MSGTYPE_AUTH_DENY:       /* authentication denied message */
        tmp = parts[0];

        if (strlen(tmp) <= UDP_MAX_MSG_LEN)
            break;

        tmp[UDP_MAX_MSG_LEN] = '\0';
        break;

    case MSGTYPE_URL:
        /* We can't truncate the URL, so let's see if we can truncate the description. */
        tmp = parts[1];
        maxlen = UDP_MAX_MSG_LEN - (strlen(tmp) + 2);
        if (maxlen < 0)
            return 1;

        if (strlen(tmp) <= maxlen)
            break;   /* no truncation needed */

        /* Truncate the description field */
        tmp[maxlen] = '\0';
        break;

    case MSGTYPE_CONTACT_ADD:     /* you were added to contact list */
    case MSGTYPE_AUTH_REQUEST:    /* authentication request */

        /* We can't truncate the user data, so let's just call it "constant" instead. */
        maxlen = UDP_MAX_MSG_LEN - (j_strlen(parts[0]) + j_strlen(parts[1]) + j_strlen(parts[2]) + j_strlen(parts[3]) + 4);

        if (m->type == MSGTYPE_AUTH_REQUEST)
        {
            /* get the "reason" parameter and truncate it */
            tmp = parts[4];

            if (j_strlen(tmp) <= maxlen)
                break;  /* we don't need to truncate */

            /* Truncate the "reason" field  */
            tmp[maxlen] = '\0';
        }
        else
        {
            /* just check to see if the packet is too long */
            if (maxlen < 0)
                return 1;
        }
        break;

    case MSGTYPE_AUTHORIZE:
        break;   /* nothing needs to get truncated here */

    case MSGTYPE_CONTACTS:
    {
        int curlen = 0;

        /* Now on to the truncate phase! */
        while (curlen > UDP_MAX_MSG_LEN)
        {
            /* XXX knock contacts off the end until we come down under the limit */
        }
        return 1;
    }
    break;

    default:
        log_debug(ZONE,"unknown message type %u",m->type);
        return 1;
    }

    return 0;
}
