/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

result it_receive(instance i, dpacket d, void *arg);
result it_sessions(void *arg);
void it_shutdown(void *arg);

void icqtrans(instance i, xmlnode x)
{
    iti ti;
    pool p = i->p;
    xmlnode config;
    char *server;

    log_debug(ZONE,"ICQ Transport, initializing for section '%s'",i->id);

    /* create new transport instance */
    ti = pmalloco(p,sizeof(_iti));
    ti->i = i;
    ti->xc = xdb_cache(i);

    config = xdb_get(ti->xc,jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:icqtrans");
    if (config == NULL)
    {
        log_error(ZONE,"ICQ Transport, configuration not found");
        return;
    }

    ti->reg = pstrdup(p,xmlnode_get_tag_data(config,"instructions"));
    if (ti->reg == NULL)
    {
        log_error(ZONE,"ICQ Transport, no instructions configured");
        xmlnode_free(config);
        return;
    }

    ti->search_msg = pstrdup(p,xmlnode_get_tag_data(config,"search"));
    if (ti->search_msg == NULL)
    {
        log_error(ZONE,"ICQ Transport, no search instructions configured");
        xmlnode_free(config);
        return;
    }

    server = xmlnode_get_tag_data(config,"server");
    ti->server = server ? strdup(server) : strdup(DEFAULT_ICQ_HOST);
    ti->sessions = ghash_create(j_atoi(xmlnode_get_tag_data(config,"prime"),509),
                                (KEYHASHFUNC) str_hash_code,(KEYCOMPAREFUNC) j_strcmp);

    ti->disable_tcp = xmlnode_get_tag(config,"disable-tcp") ? 1 : 0;
    if (ti->disable_tcp)
    {
        log_debug(ZONE,"TCP disabled");
    }
    else
    {
        ti->port_high = j_atoi(xmlnode_get_tag_data(config,"ports/max"),3000);
        ti->port = ti->port_low = j_atoi(xmlnode_get_tag_data(config,"ports/min"),2000);
    }

    ti->vcard = xmlnode_new_tag_pool(p,"vCard");
    xmlnode_put_attrib(ti->vcard,"xmlns",NS_VCARD);
    xmlnode_insert_node(ti->vcard,xmlnode_get_firstchild(xmlnode_get_tag(config,"vCard")));

    ti->start = time(NULL);

    xmlnode_free(config);

    it_check_endian();

    register_phandler(i,o_DELIVER,it_receive,(void *) ti);
    register_beat(1,it_sessions,(void *) ti);
    register_shutdown(it_shutdown,(void *) ti);
}

result it_receive(instance i, dpacket d, void *arg)
{
    iti ti = (iti) arg;
    jpacket jp;
    session s;

    jp = jpacket_new(d->x);

    if (!jp->from || !jp->from->user || jpacket_subtype(jp) == JPACKET__ERROR || ti->shutdown == 1)
    {
        log_debug(ZONE,"ingoring packet");
        xmlnode_free(jp->x);
        return r_DONE;
    }

    s = (session) ghash_get(ti->sessions,jid_full(jid_user(jp->from)));
    if (s != NULL && s->exit_flag == 0)
        it_session_jpacket(s,jp);
    else
        it_unknown(ti,jp);

    return r_DONE;
}

result it_sessions(void *arg)
{
    iti ti = (iti) arg;
    session cur;

    for (cur = ti->session_list; cur != NULL; cur = cur->next)
    {
        if (cur->exit_flag) continue;

        if (cur->rsend != -1 && time(NULL) > cur->rsend)
            it_server_retransmit(cur);

        if (cur->connected)
        {
            if (cur->timer + KEEPALIVE < time(NULL))
                it_server_keepalive(cur);
        }
        else
        {
            if (cur->timer + LOGIN_TIMER < time(NULL))
            {
                log_debug(ZONE,"Session login timeout");
                if (cur->type != stype_normal)
                    it_session_register_fail(cur,TERROR_REQTIMEOUT);
                it_session_end(cur);
            }
        }
    }

    if (ti->shutdown && ti->sessions_count == 0 && ti->shutdown_cond)
        pth_cond_notify(ti->shutdown_cond,FALSE);

    return r_DONE;
}

int it_sessions_end(void *arg, const void *key, void *data)
{
    it_session_end((session) data);
    return 1;
}

void it_shutdown(void *arg)
{
    iti ti = (iti) arg;
    pth_mutex_t mutex = PTH_MUTEX_INIT;
    pth_cond_t cond = PTH_COND_INIT;
    log_notice(ZONE,"ICQ Transport, shutting down");

    ti->shutdown = 1;

    ghash_walk(ti->sessions,it_sessions_end,(void *) ti);

    /* wait for the sessions to end */
    ti->shutdown_cond = &cond;
    pth_mutex_acquire(&mutex,FALSE,NULL);
    pth_cond_await(&cond,&mutex,NULL);

    ghash_destroy(ti->sessions);
    free(ti->server);
}
