/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 */
#include "icqtransport.h"

void it_server_read(mio m, int state, void *arg, char *buffer, int len);
int it_decode_inbound_packet(session s, void *buffer, unsigned len, PICQDECODE pdecode);

int it_server_connect(session s)
{
    iti ti = s->ti;
    struct sockaddr_in addr;
    int rc, sock;

    log_debug(ZONE,"server_connect");

    memset(&addr,0,sizeof(struct sockaddr_in));
    if (inet_aton(ti->server,&(addr.sin_addr)) == 0)
        return 1;

    /* Finish filling in the IP address structure. */
    addr.sin_family = AF_INET;
    addr.sin_port = htons((unsigned short)(DEFAULT_ICQ_PORT));

    /* Create the UDP socket that will connect to the server. */
    sock = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
    if (sock == -1)
    {
        /* error creating the socket */
        log_error(ZONE,"server_connect: error creating server socket: %s",strerror(errno));
        return 1;
    }

    /* Connect that UDP socket to the server. */
    if (pth_connect(sock,(struct sockaddr *)(&addr),sizeof(struct sockaddr_in)) == -1)
    {
        /* connect failed, need to close the socket */
        log_error(ZONE,"server_connect: error connecting server socket: %s",strerror(errno));
        close(sock);
        return 1;
    }

    /* Get the local address our socket is bound to. */
    rc = sizeof(struct sockaddr_in);
    getsockname(sock,(struct sockaddr *)(&(s->addr)),&rc);
    log_debug(ZONE,"- our local IP address is %s",inet_ntoa(s->addr.sin_addr));

    s->server = mio_new(sock,it_server_read,(void *) s,mio_handlers_new(NULL,NULL,NULL));
    mio_karma(s->server,KARMA_INIT,KARMA_MAX,KARMA_INC,0,KARMA_PENALTY,KARMA_RESTORE);

    return 0;
}

int it_server_repeated(session s, icqshort seq)
{
    int i;

    for (i = 0; i < RSEQ_MAX; i++)
        if (s->rseq[i] == seq)
        {
            log_debug(ZONE,"%04X:%04X",s->rseq[i],seq);
            return 1;
        }

    return 0;
}

void it_server_ack(session s, unsigned short seq1, unsigned short seq2)
{
    ICQBUFFER5C packet;
    unsigned long check;

    /* Initialize the packet and fill in the fields. */
    memset(&(packet.hdr),0,sizeof(ICQHDR5C));
    packet.hdr.version = to_icqshort(ICQ_VER_5);
    packet.hdr.uin = to_icqlong(s->uin);
    packet.hdr.sid = to_icqlong(s->sid);
    packet.hdr.command = to_icqshort(CMD_ACK);
    packet.hdr.seq1 = to_icqshort(seq1);
    packet.hdr.seq2 = to_icqshort(seq2);
    put_icqlong(packet.data,0,random());

    /* Compute the checkvalue and add it to the packet. */
    check = it_checkvalue5((const icqbyte *)&packet,sizeof(ICQHDR5C) + 4);
    packet.hdr.checkvalue = to_icqlong(check);

    /* Now "encrypt" the packet. */
    it_scramble5(&packet,sizeof(ICQHDR5C) + 4,check);

    it_server_write(s,&packet,sizeof(ICQHDR5C) + 4,CMD_ACK,seq1);
}

void it_server_read(mio m, int state, void *arg, char *buffer, int len)
{
    session s = (session) arg;
    ICQDECODE packet;

    switch(state)
    {
    case MIO_CLOSED:
        log_debug(ZONE,"MIO closed");
        if (s)
        {
            s->server = NULL;
            it_session_end(s);
            /* XXX reconnect */
        }
        return;
    case MIO_BUFFER:
        break;
    default:
        return;
    }

    if (s == NULL)
        return;

    if (it_decode_inbound_packet(s,buffer,(unsigned) len,&packet))
    {
        if (packet.seq1 != 0 && it_server_repeated(s,packet.seq1))
        {
            log_debug(ZONE,"Repeated packet %04X",packet.command);

            if (packet.command != SRV_ACK)
                it_server_ack(s,packet.seq1,packet.seq2);

            return;
        }

        if (packet.command != SRV_ACK)
        {
#ifdef PACKET_DEBUG
            it_debug_dump(ZONE,buffer,len);
#endif
            it_server_ack(s,packet.seq1,packet.seq2);

            /* save this sequence number in the list of received numbers */
            s->rseq[s->rseq_cur++] = packet.seq1;
            if (s->rseq_cur == RSEQ_MAX)
                s->rseq_cur = 0;

            it_server_process_packet(s,&packet);
        }
        else
            it_server_acked_packet(s,packet.seq1);
    }
    else
        mio_close(m); /* packet parsing error occurred */
}

int it_decode_inbound_packet(session s, void *buffer, unsigned len, PICQDECODE pdecode)
{
    int rc;

    memset(pdecode,0,sizeof(ICQDECODE));
    if (len<sizeof(icqshort))
    {
        log_debug(ZONE,"dropped packet (bad length) %u",len);
        return 0;
    }

    pdecode->version = get_icqshort(buffer,0);
    rc = it_test_v5_server(s,buffer,len);
    if (rc)
    {
        /* decode the header stuff */
        PICQHDR5S hdr = (PICQHDR5S)buffer;
        pdecode->command = from_icqshort(hdr->command);
        pdecode->seq1 = from_icqshort(hdr->seq1);
        pdecode->seq2 = from_icqshort(hdr->seq2);
        pdecode->uin = from_icqlong(hdr->uin);
        pdecode->data = (icqbyte *)buffer;
        pdecode->data += sizeof(ICQHDR5S);
        pdecode->datalen = len - sizeof(ICQHDR5S);
    }

    return rc;
}

void it_server_write(session s, void *buf, unsigned len, unsigned short cmd,  unsigned short seq1)
{
    if (cmd != CMD_ACK && s->exit_flag == 0)
    {
        pool p;
        pqueue pq;

        log_debug(ZONE,"Session[%s], server write %04X, length %d",jid_full(s->id),seq1,(int) len);

        p = pool_new();
        pq = pmalloco(p,sizeof(_pqueue));
        pq->p = p;

        pq->exp_time = time(NULL) + RESEND_DELAY;
        pq->seq1 = seq1;
        pq->len = (unsigned short) len;
        pq->packet = pmalloco(p,len);
        memcpy(pq->packet,buf,len);

        if (s->pq_tail)
        {
            s->pq_tail->next = pq;
            s->pq_tail = pq;
        }
        else
            s->pq_head = s->pq_tail = pq;

        s->rsend = s->pq_head->exp_time;
    }

    mio_write(s->server,NULL,(char *) buf,len);
}
/* prepares the header of a version 5 packet and scambles it */
unsigned short it_prepare_packet(session s, void *packet, unsigned len, unsigned short command)
{
    PICQHDR5C hdr = (PICQHDR5C)packet;
    unsigned long check;
    unsigned short rc;

    /* Initialize the header and fill in the "known" fields. */
    memset(hdr,0,sizeof(ICQHDR5C));
    hdr->version = to_icqshort(ICQ_VER_5);
    hdr->uin = to_icqlong(s->uin);
    hdr->sid = to_icqlong(s->sid);
    hdr->command = to_icqshort(command);
    rc = s->seq1++;
    hdr->seq1 = to_icqshort(rc);
    if ((command!=CMD_KEEP_ALIVE) && (command!=CMD_SEND_TEXT_CODE)
           && (command!=CMD_KEEP_ALIVE2))
        hdr->seq2 = to_icqshort(s->seq2++);

    /* Compute the checkvalue and add it to the packet. */
    check = it_checkvalue5((const icqbyte *)packet,len);
    hdr->checkvalue = to_icqlong(check);

#ifdef PACKET_DEBUG
    it_debug_dump(ZONE,packet,len);
#endif

    /* Now encrypt the packet. */
    it_scramble5(packet,len,check);

    return rc; 
}

void it_build_and_write(session s, PICQBUFFER5C packet, unsigned len, unsigned short cmd)
{
    unsigned packet_len = len + sizeof(ICQHDR5C);
    unsigned short seq1;

    seq1 = it_prepare_packet(s,packet,packet_len,cmd);
    it_server_write(s,packet,packet_len,cmd,seq1);    
}

void it_server_acked_packet(session s, unsigned short seq1)
{
    pqueue cur, prev;

    for (prev = NULL, cur = s->pq_head; cur != NULL; prev = cur, cur = cur->next)
    {
        if (cur->seq1 != seq1) continue;

        log_debug(ZONE,"Session[%s], server acked %04X",jid_full(s->id),seq1);

        if (prev)
        {
            if (s->pq_tail == cur)
            {
                prev->next = NULL;
                s->pq_tail = prev;
            }
            else
                prev->next = cur->next;
        }
        else
        {
            if (cur == s->pq_tail)
                s->pq_head = s->pq_tail = NULL;
            else
                s->pq_head = cur->next;
        }

        if (s->pq_head)
            s->rsend = s->pq_head->exp_time;
        else
            s->rsend = -1;

        pool_free(cur->p);
        return;
    }

    log_debug(ZONE,"Received duped ACK, %04X",seq1);
}

void it_server_retransmit(session s)
{
    pqueue top = s->pq_head;

    if (++top->attempts == MAX_RETRY_ATTEMPTS)
    {
        log_debug(ZONE,"Session[%s], disgarding packet %04X",jid_full(s->id),top->seq1);

        if (s->pq_tail == top)
            s->pq_head = s->pq_tail = NULL;
        else
            s->pq_head = top->next;

        pool_free(top->p);
        top = s->pq_head;
    }
    else
    {
        log_debug(ZONE,"Session[%s], retransmit %04X, attempt %d",jid_full(s->id),top->seq1,top->attempts);

        top->exp_time = time(NULL) + RESEND_DELAY;

        if (s->pq_tail != top)
        {
            /* move it to the end */
            s->pq_head = top->next;
            top->next = NULL;

            s->pq_tail->next = top;
            s->pq_tail = top;
        }

        mio_write(s->server,NULL,(char *) top->packet, (int) top->len);
    }

    if (s->pq_head)
        s->rsend = s->pq_head->exp_time;
    else
        s->rsend = -1; /* noting to resend at the moment */
}
